/*
 ---------------------------------------------------------------------------
 Copyright (c) 2002, Dr Brian Gladman, Worcester, UK.   All rights reserved.

 LICENSE TERMS

 The free distribution and use of this software in both source and binary
 form is allowed (with or without changes) provided that:

   1. distributions of this source code include the above copyright
      notice, this list of conditions and the following disclaimer;

   2. distributions in binary form include the above copyright
      notice, this list of conditions and the following disclaimer
      in the documentation and/or other associated materials;

   3. the copyright holder's name is not used to endorse products
      built using this software without specific written permission.

 ALTERNATIVELY, provided that this notice is retained in full, this product
 may be distributed under the terms of the GNU General Public License (GPL),
 in which case the provisions of the GPL apply INSTEAD OF those given above.

 DISCLAIMER

 This software is provided 'as is' with no explicit or implied warranties
 in respect of its properties, including, but not limited to, correctness
 and/or fitness for purpose.
 ---------------------------------------------------------------------------
 Issue Date: 18/06/2004

 This is a byte oriented version of SHA1 that operates on arrays of bytes
 stored in memory.
*/

#include <string.h>     /* for memcpy() etc.        */
#include <stdlib.h>     /* for _lrotl with VC++     */

#include "Sha1.h"

/*
    To obtain the highest speed on processors with 32-bit words, this code
    needs to determine the order in which bytes are packed into such words.
    The following block of code is an attempt to capture the most obvious
    ways in which various environemnts specify their endian definitions.
    It may well fail, in which case the definitions will need to be set by
    editing at the points marked **** EDIT HERE IF NECESSARY **** below.
*/

/*  PLATFORM SPECIFIC INCLUDES */

/* Original byte order detection removed */
//#include "../Common/Endian.h"

#define BRG_LITTLE_ENDIAN   1234 /* byte 0 is least significant (i386) */
#define BRG_BIG_ENDIAN      4321 /* byte 0 is most significant (mc68k) */

#if BYTE_ORDER == LITTLE_ENDIAN
#  define PLATFORM_BYTE_ORDER BRG_LITTLE_ENDIAN
#endif
/*
#if BYTE_ORDER == BIG_ENDIAN
#  define PLATFORM_BYTE_ORDER BRG_BIG_ENDIAN
#endif
/**/
#ifdef _MSC_VER
#pragma intrinsic(memcpy)
#endif

#if 1 && defined(_MSC_VER) && !defined(_DEBUG)
#define rotl32  _rotl
#define rotr32  _rotr
#else
#define rotl32(x,n)   (((x) << n) | ((x) >> (32 - n)))
#define rotr32(x,n)   (((x) >> n) | ((x) << (32 - n)))
#endif

#if !defined(bswap_32)
#define bswap_32(x) (rotr32((x), 24) & 0x00ff00ff | rotr32((x), 8) & 0xff00ff00)
#endif

#if (PLATFORM_BYTE_ORDER == BRG_LITTLE_ENDIAN)
#define SWAP_BYTES
#else
#undef  SWAP_BYTES
#endif

#if defined(SWAP_BYTES)
#define bsw_32(p,n) \
    { int _i = (n); while(_i--) ((sha1_32t*)p)[_i] = bswap_32(((sha1_32t*)p)[_i]); }
#else
#define bsw_32(p,n)
#endif

#define SHA1_MASK   (SHA1_BLOCK_SIZE - 1)

#if 0

#define ch(x,y,z)       (((x) & (y)) ^ (~(x) & (z)))
#define parity(x,y,z)   ((x) ^ (y) ^ (z))
#define maj(x,y,z)      (((x) & (y)) ^ ((x) & (z)) ^ ((y) & (z)))

#else   /* Discovered by Rich Schroeppel and Colin Plumb   */

#define ch(x,y,z)       ((z) ^ ((x) & ((y) ^ (z))))
#define parity(x,y,z)   ((x) ^ (y) ^ (z))
#define maj(x,y,z)      (((x) & (y)) | ((z) & ((x) ^ (y))))

#endif

/* Compile 64 bytes of hash data into SHA1 context. Note    */
/* that this routine assumes that the byte order in the     */
/* ctx->wbuf[] at this point is in such an order that low   */
/* address bytes in the ORIGINAL byte stream will go in     */
/* this buffer to the high end of 32-bit words on BOTH big  */
/* and little endian systems                                */

#ifdef ARRAY
#define q(v,n)  v[n]
#else
#define q(v,n)  v##n
#endif

#define one_cycle(v,a,b,c,d,e,f,k,h)            \
    q(v,e) += rotr32(q(v,a),27) +               \
              f(q(v,b),q(v,c),q(v,d)) + k + h;  \
    q(v,b)  = rotr32(q(v,b), 2)

#define five_cycle(v,f,k,i)                 \
    one_cycle(v, 0,1,2,3,4, f,k,hf(i  ));   \
    one_cycle(v, 4,0,1,2,3, f,k,hf(i+1));   \
    one_cycle(v, 3,4,0,1,2, f,k,hf(i+2));   \
    one_cycle(v, 2,3,4,0,1, f,k,hf(i+3));   \
    one_cycle(v, 1,2,3,4,0, f,k,hf(i+4))


void sha1_compile(sha1_ctx ctx[1], int rounds)
{   sha1_32t    *w = ctx->wbuf;

#ifdef ARRAY
    sha1_32t    v[5];
    memcpy(v, ctx->hash, 5 * sizeof(sha1_32t));
#else
    sha1_32t    v0, v1, v2, v3, v4;
    v0 = ctx->hash[0]; v1 = ctx->hash[1];
    v2 = ctx->hash[2]; v3 = ctx->hash[3];
    v4 = ctx->hash[4];
#endif

#define hf(i)   w[i]

    //five_cycle(v, ch, 0x5a827999,  0);
    if (rounds >= 1) one_cycle(v, 0,1,2,3,4, ch,0x5a827999,hf(0  ));   
    if (rounds >= 2) one_cycle(v, 4,0,1,2,3, ch,0x5a827999,hf(0+1));   
    if (rounds >= 3) one_cycle(v, 3,4,0,1,2, ch,0x5a827999,hf(0+2));   
    if (rounds >= 4) one_cycle(v, 2,3,4,0,1, ch,0x5a827999,hf(0+3));   
    if (rounds >= 5) one_cycle(v, 1,2,3,4,0, ch,0x5a827999,hf(0+4));
    
    //five_cycle(v, ch, 0x5a827999,  5);
    if (rounds >= 6) one_cycle(v, 0,1,2,3,4, ch,0x5a827999,hf(5  ));   
    if (rounds >= 7) one_cycle(v, 4,0,1,2,3, ch,0x5a827999,hf(5+1));   
    if (rounds >= 8) one_cycle(v, 3,4,0,1,2, ch,0x5a827999,hf(5+2));   
    if (rounds >= 9) one_cycle(v, 2,3,4,0,1, ch,0x5a827999,hf(5+3));   
    if (rounds >= 10) one_cycle(v, 1,2,3,4,0, ch,0x5a827999,hf(5+4));

    //five_cycle(v, ch, 0x5a827999, 10);
    if (rounds >= 11) one_cycle(v, 0,1,2,3,4, ch,0x5a827999,hf(10  ));   
    if (rounds >= 12) one_cycle(v, 4,0,1,2,3, ch,0x5a827999,hf(10+1));   
    if (rounds >= 13) one_cycle(v, 3,4,0,1,2, ch,0x5a827999,hf(10+2));   
    if (rounds >= 14) one_cycle(v, 2,3,4,0,1, ch,0x5a827999,hf(10+3));   
    if (rounds >= 15) one_cycle(v, 1,2,3,4,0, ch,0x5a827999,hf(10+4));
    
    if (rounds >= 16) one_cycle(v,0,1,2,3,4, ch, 0x5a827999, hf(15)); 

#undef  hf
#define hf(i) (w[(i) & 15] = rotl32(                    \
                 w[((i) + 13) & 15] ^ w[((i) + 8) & 15] \
               ^ w[((i) +  2) & 15] ^ w[(i) & 15], 1))

    if (rounds >= 17) one_cycle(v,4,0,1,2,3, ch, 0x5a827999, hf(16));
    if (rounds >= 18) one_cycle(v,3,4,0,1,2, ch, 0x5a827999, hf(17));
    if (rounds >= 19) one_cycle(v,2,3,4,0,1, ch, 0x5a827999, hf(18));
    if (rounds >= 20) one_cycle(v,1,2,3,4,0, ch, 0x5a827999, hf(19));

    //five_cycle(v, parity, 0x6ed9eba1,  20);
    if (rounds >= 21) one_cycle(v, 0,1,2,3,4, parity,0x6ed9eba1,hf(20  ));   
    if (rounds >= 22) one_cycle(v, 4,0,1,2,3, parity,0x6ed9eba1,hf(20+1));   
    if (rounds >= 23) one_cycle(v, 3,4,0,1,2, parity,0x6ed9eba1,hf(20+2));   
    if (rounds >= 24) one_cycle(v, 2,3,4,0,1, parity,0x6ed9eba1,hf(20+3));   
    if (rounds >= 25) one_cycle(v, 1,2,3,4,0, parity,0x6ed9eba1,hf(20+4));

    //five_cycle(v, parity, 0x6ed9eba1,  25);
    if (rounds >= 26) one_cycle(v, 0,1,2,3,4, parity,0x6ed9eba1,hf(25  ));   
    if (rounds >= 27) one_cycle(v, 4,0,1,2,3, parity,0x6ed9eba1,hf(25+1));   
    if (rounds >= 28) one_cycle(v, 3,4,0,1,2, parity,0x6ed9eba1,hf(25+2));   
    if (rounds >= 29) one_cycle(v, 2,3,4,0,1, parity,0x6ed9eba1,hf(25+3));   
    if (rounds >= 30) one_cycle(v, 1,2,3,4,0, parity,0x6ed9eba1,hf(25+4));
    
    //five_cycle(v, parity, 0x6ed9eba1,  30);
    if (rounds >= 31) one_cycle(v, 0,1,2,3,4, parity,0x6ed9eba1,hf(30  ));   
    if (rounds >= 32) one_cycle(v, 4,0,1,2,3, parity,0x6ed9eba1,hf(30+1));   
    if (rounds >= 33) one_cycle(v, 3,4,0,1,2, parity,0x6ed9eba1,hf(30+2));   
    if (rounds >= 34) one_cycle(v, 2,3,4,0,1, parity,0x6ed9eba1,hf(30+3));   
    if (rounds >= 35) one_cycle(v, 1,2,3,4,0, parity,0x6ed9eba1,hf(30+4));
    
    //five_cycle(v, parity, 0x6ed9eba1,  35);
    if (rounds >= 36) one_cycle(v, 0,1,2,3,4, parity,0x6ed9eba1,hf(35  ));   
    if (rounds >= 37) one_cycle(v, 4,0,1,2,3, parity,0x6ed9eba1,hf(35+1));   
    if (rounds >= 38) one_cycle(v, 3,4,0,1,2, parity,0x6ed9eba1,hf(35+2));   
    if (rounds >= 39) one_cycle(v, 2,3,4,0,1, parity,0x6ed9eba1,hf(35+3));   
    if (rounds >= 40) one_cycle(v, 1,2,3,4,0, parity,0x6ed9eba1,hf(35+4));

    //five_cycle(v, maj, 0x8f1bbcdc,  40);
    if (rounds >= 41) one_cycle(v, 0,1,2,3,4, maj,0x8f1bbcdc,hf(40  ));   
    if (rounds >= 42) one_cycle(v, 4,0,1,2,3, maj,0x8f1bbcdc,hf(40+1));   
    if (rounds >= 43) one_cycle(v, 3,4,0,1,2, maj,0x8f1bbcdc,hf(40+2));   
    if (rounds >= 44) one_cycle(v, 2,3,4,0,1, maj,0x8f1bbcdc,hf(40+3));   
    if (rounds >= 45) one_cycle(v, 1,2,3,4,0, maj,0x8f1bbcdc,hf(40+4));
    
    //five_cycle(v, maj, 0x8f1bbcdc,  45);
    if (rounds >= 46) one_cycle(v, 0,1,2,3,4, maj,0x8f1bbcdc,hf(45  ));   
    if (rounds >= 47) one_cycle(v, 4,0,1,2,3, maj,0x8f1bbcdc,hf(45+1));   
    if (rounds >= 48) one_cycle(v, 3,4,0,1,2, maj,0x8f1bbcdc,hf(45+2));   
    if (rounds >= 49) one_cycle(v, 2,3,4,0,1, maj,0x8f1bbcdc,hf(45+3));   
    if (rounds >= 50) one_cycle(v, 1,2,3,4,0, maj,0x8f1bbcdc,hf(45+4));
    
    //five_cycle(v, maj, 0x8f1bbcdc,  50);
    if (rounds >= 51) one_cycle(v, 0,1,2,3,4, maj,0x8f1bbcdc,hf(50  ));   
    if (rounds >= 52) one_cycle(v, 4,0,1,2,3, maj,0x8f1bbcdc,hf(50+1));   
    if (rounds >= 53) one_cycle(v, 3,4,0,1,2, maj,0x8f1bbcdc,hf(50+2));   
    if (rounds >= 54) one_cycle(v, 2,3,4,0,1, maj,0x8f1bbcdc,hf(50+3));   
    if (rounds >= 55) one_cycle(v, 1,2,3,4,0, maj,0x8f1bbcdc,hf(50+4));
    
    //five_cycle(v, maj, 0x8f1bbcdc,  55);
    if (rounds >= 56) one_cycle(v, 0,1,2,3,4, maj,0x8f1bbcdc,hf(55  ));   
    if (rounds >= 57) one_cycle(v, 4,0,1,2,3, maj,0x8f1bbcdc,hf(55+1));   
    if (rounds >= 58) one_cycle(v, 3,4,0,1,2, maj,0x8f1bbcdc,hf(55+2));   
    if (rounds >= 59) one_cycle(v, 2,3,4,0,1, maj,0x8f1bbcdc,hf(55+3));   
    if (rounds >= 60) one_cycle(v, 1,2,3,4,0, maj,0x8f1bbcdc,hf(55+4));

    //five_cycle(v, parity, 0xca62c1d6,  60);
    if (rounds >= 61) one_cycle(v, 0,1,2,3,4, parity,0xca62c1d6,hf(60  ));   
    if (rounds >= 62) one_cycle(v, 4,0,1,2,3, parity,0xca62c1d6,hf(60+1));   
    if (rounds >= 63) one_cycle(v, 3,4,0,1,2, parity,0xca62c1d6,hf(60+2));   
    if (rounds >= 64) one_cycle(v, 2,3,4,0,1, parity,0xca62c1d6,hf(60+3));   
    if (rounds >= 65) one_cycle(v, 1,2,3,4,0, parity,0xca62c1d6,hf(60+4));
    
    //five_cycle(v, parity, 0xca62c1d6,  65);
    if (rounds >= 66) one_cycle(v, 0,1,2,3,4, parity,0xca62c1d6,hf(65  ));   
    if (rounds >= 67) one_cycle(v, 4,0,1,2,3, parity,0xca62c1d6,hf(65+1));   
    if (rounds >= 68) one_cycle(v, 3,4,0,1,2, parity,0xca62c1d6,hf(65+2));   
    if (rounds >= 69) one_cycle(v, 2,3,4,0,1, parity,0xca62c1d6,hf(65+3));   
    if (rounds >= 70) one_cycle(v, 1,2,3,4,0, parity,0xca62c1d6,hf(65+4));
    
    //five_cycle(v, parity, 0xca62c1d6,  70);
    if (rounds >= 71) one_cycle(v, 0,1,2,3,4, parity,0xca62c1d6,hf(70  ));   
    if (rounds >= 72) one_cycle(v, 4,0,1,2,3, parity,0xca62c1d6,hf(70+1));   
    if (rounds >= 73) one_cycle(v, 3,4,0,1,2, parity,0xca62c1d6,hf(70+2));   
    if (rounds >= 74) one_cycle(v, 2,3,4,0,1, parity,0xca62c1d6,hf(70+3));   
    if (rounds >= 75) one_cycle(v, 1,2,3,4,0, parity,0xca62c1d6,hf(70+4));
    
    //five_cycle(v, parity, 0xca62c1d6,  75);
    if (rounds >= 76) one_cycle(v, 0,1,2,3,4, parity,0xca62c1d6,hf(75  ));   
    if (rounds >= 77) one_cycle(v, 4,0,1,2,3, parity,0xca62c1d6,hf(75+1));   
    if (rounds >= 78) one_cycle(v, 3,4,0,1,2, parity,0xca62c1d6,hf(75+2));   
    if (rounds >= 79) one_cycle(v, 2,3,4,0,1, parity,0xca62c1d6,hf(75+3));   
    if (rounds >= 80) one_cycle(v, 1,2,3,4,0, parity,0xca62c1d6,hf(75+4));
    

#ifdef ARRAY
    ctx->hash[0] += v[0]; ctx->hash[1] += v[1];
    ctx->hash[2] += v[2]; ctx->hash[3] += v[3];
    ctx->hash[4] += v[4];
#else
    ctx->hash[0] += v0; ctx->hash[1] += v1;
    ctx->hash[2] += v2; ctx->hash[3] += v3;
    ctx->hash[4] += v4;
#endif
}

/**/
/*
void sha1_compile(sha1_ctx ctx[1], int rounds)
{   sha1_32t    *w = ctx->wbuf;

#ifdef ARRAY
    sha1_32t    v[5];
    memcpy(v, ctx->hash, 5 * sizeof(sha1_32t));
#else
    sha1_32t    v0, v1, v2, v3, v4;
    v0 = ctx->hash[0]; v1 = ctx->hash[1];
    v2 = ctx->hash[2]; v3 = ctx->hash[3];
    v4 = ctx->hash[4];
#endif

#define hf(i)   w[i]


    five_cycle(v, ch, 0x5a827999,  0);
    five_cycle(v, ch, 0x5a827999,  5);
    five_cycle(v, ch, 0x5a827999, 10);
    
    one_cycle(v,0,1,2,3,4, ch, 0x5a827999, hf(15)); \

#undef  hf
#define hf(i) (w[(i) & 15] = rotl32(                    \
                 w[((i) + 13) & 15] ^ w[((i) + 8) & 15] \
               ^ w[((i) +  2) & 15] ^ w[(i) & 15], 1))

    one_cycle(v,4,0,1,2,3, ch, 0x5a827999, hf(16));
    one_cycle(v,3,4,0,1,2, ch, 0x5a827999, hf(17));
    one_cycle(v,2,3,4,0,1, ch, 0x5a827999, hf(18));
    one_cycle(v,1,2,3,4,0, ch, 0x5a827999, hf(19));

    five_cycle(v, parity, 0x6ed9eba1,  20);
    five_cycle(v, parity, 0x6ed9eba1,  25);
    five_cycle(v, parity, 0x6ed9eba1,  30);
    five_cycle(v, parity, 0x6ed9eba1,  35);

    five_cycle(v, maj, 0x8f1bbcdc,  40);
    five_cycle(v, maj, 0x8f1bbcdc,  45);
    five_cycle(v, maj, 0x8f1bbcdc,  50);
    five_cycle(v, maj, 0x8f1bbcdc,  55);

    five_cycle(v, parity, 0xca62c1d6,  60);
    five_cycle(v, parity, 0xca62c1d6,  65);
    five_cycle(v, parity, 0xca62c1d6,  70);
    five_cycle(v, parity, 0xca62c1d6,  75);

#ifdef ARRAY
    ctx->hash[0] += v[0]; ctx->hash[1] += v[1];
    ctx->hash[2] += v[2]; ctx->hash[3] += v[3];
    ctx->hash[4] += v[4];
#else
    ctx->hash[0] += v0; ctx->hash[1] += v1;
    ctx->hash[2] += v2; ctx->hash[3] += v3;
    ctx->hash[4] += v4;
#endif
}
/**/
void sha1_begin(sha1_ctx* ctx)
{
    ctx->count[0] = ctx->count[1] = 0;
    ctx->hash[0] = 0x67452301;
    ctx->hash[1] = 0xefcdab89;
    ctx->hash[2] = 0x98badcfe;
    ctx->hash[3] = 0x10325476;
    ctx->hash[4] = 0xc3d2e1f0;
}

/* SHA1 hash data in an array of bytes into hash buffer and */
/* call the hash_compile function as required.              */

void sha1_hash(sha1_ctx* ctx, const unsigned char data[], unsigned long len, int rounds) {
   sha1_32t pos = (sha1_32t)(ctx->count[0] & SHA1_MASK),
            space = SHA1_BLOCK_SIZE - pos;
    const unsigned char *sp = data;

	if (rounds == -1) rounds = SHA1_ROUNDS;

    if((ctx->count[0] += len) < len)
        ++(ctx->count[1]);

    while(len >= space)     /* tranfer whole blocks if possible  */
    {
        memcpy(((unsigned char*)ctx->wbuf) + pos, sp, space);
        sp += space; len -= space; space = SHA1_BLOCK_SIZE; pos = 0;
        bsw_32(ctx->wbuf, SHA1_BLOCK_SIZE >> 2);
        sha1_compile(ctx, rounds);
    }

    memcpy(((unsigned char*)ctx->wbuf) + pos, sp, len);
}

/* SHA1 final padding and digest calculation  */

void sha1_end(sha1_ctx* ctx, unsigned char hval[], int rounds)
{   sha1_32t    i = (sha1_32t)(ctx->count[0] & SHA1_MASK);

	if (rounds == -1) rounds = SHA1_ROUNDS;

    /* put bytes in the buffer in an order in which references to   */
    /* 32-bit words will put bytes with lower addresses into the    */
    /* top of 32 bit words on BOTH big and little endian machines   */
    bsw_32(ctx->wbuf, (i + 3) >> 2);

    /* we now need to mask valid bytes and add the padding which is */
    /* a single 1 bit and as many zero bits as necessary. Note that */
    /* we can always add the first padding byte here because the    */
    /* buffer always has at least one empty slot                    */
    ctx->wbuf[i >> 2] &= 0xffffff80 << 8 * (~i & 3);
    ctx->wbuf[i >> 2] |= 0x00000080 << 8 * (~i & 3);

    /* we need 9 or more empty positions, one for the padding byte  */
    /* (above) and eight for the length count. If there is not      */
    /* enough space, pad and empty the buffer                       */
    if(i > SHA1_BLOCK_SIZE - 9)
    {
        if(i < 60) ctx->wbuf[15] = 0;
        sha1_compile(ctx, rounds);
        i = 0;
    }
    else    /* compute a word index for the empty buffer positions  */
        i = (i >> 2) + 1;

    while(i < 14) /* and zero pad all but last two positions        */
        ctx->wbuf[i++] = 0;

    /* the following 32-bit length fields are assembled in the      */
    /* wrong byte order on little endian machines but this is       */
    /* corrected later since they are only ever used as 32-bit      */
    /* word values.                                                 */
    ctx->wbuf[14] = (ctx->count[1] << 3) | (ctx->count[0] >> 29);
    ctx->wbuf[15] = ctx->count[0] << 3;
    sha1_compile(ctx, rounds);

    /* extract the hash value as bytes in case the hash buffer is   */
    /* misaligned for 32-bit words                                  */
    for(i = 0; i < SHA1_DIGEST_SIZE; ++i)
        hval[i] = (unsigned char)(ctx->hash[i >> 2] >> (8 * (~i & 3)));
}

void sha1(unsigned char hval[], const unsigned char data[], unsigned __int32 len, int rounds)
{   sha1_ctx    cx[1];

	if (rounds == -1) rounds = SHA1_ROUNDS;
	
    sha1_begin(cx); sha1_hash(cx, data, len, rounds); sha1_end(cx, hval, rounds);
}

